# -*- coding: utf-8 -*-

"""
ReactionSystem is a Python library for Reaction System managing.
"""

import sys
if not (sys.version_info[0] == 3 and sys.version_info[1] >= 3):
    raise ImportError("Python version 3.3 or above is required.")
del sys

try:
    import pyeda
except ImportError:
    raise ImportError("ReactionSystem depends on pyeda as an external library. ")
del pyeda

import sys
if sys.version_info[0] < 2:
    raise ImportError("Python version 3 for ReactionSystem.")
del sys

from .reaction_system import *
from .exceptions import *



class _fbs_iterate_item():
    def __init__(self, formula, parent, step, inv_nf):
        self.formula = formula
        self.parent = parent
        self.step = step
        self.inv_nf = inv_nf
        self.remained = 1
        self.childs = []


# -*- coding: utf-8 -*-

from pyeda.inter import expr
from pyeda.inter import espresso_exprs
from pyeda.boolalg.expr import Atom 

from .boolean_wrap import (
    Not,
    And,
    Or,
    Constant,
    Literal,
    Variable,
    Complement,
    AndOp,
    OrOp,
    OrAndOp,
    ZERO,
    ONE,
    var)

from .reaction import Reaction
from .reaction_set import ReactionSet
from .exceptions import ExceptionReactionSystem

from ._fbs_iterate_item import _fbs_iterate_item

import sys
if 'time' in sys.argv:
    import time


class ReactionSystem():
    def __init__(self, A):
        reactant_set = set()
        for reaction in A:
            reactant_set = reactant_set.union(reaction.P)
        self._causes = {}
        for reactant in reactant_set:
            self._causes[reactant] = A.cause(reactant)

    def cause(self, symbol):
        return self._causes.get(symbol, ZERO)

    def fbp(self, symbols, steps, context_given=set(), context_not_given=set()):
        symbolSet = Reaction._create_symbol_set(symbols)
        if not isinstance(steps, int) or steps < 0:
                raise ExceptionReactionSystem.InvalidNumber()
        if (not isinstance(context_given, set) or 
            not isinstance(context_not_given, set)):
                raise ExceptionReactionSystem.InvalidContextSet()
        
        self.cg = context_given
        self.cng = context_not_given

        if 'time' in sys.argv:
            start = time.time()

        formula = ONE
        for symbol in symbolSet:
            formula = And(formula, self.cause(symbol))
            formula = formula.to_dnf()
        
        formula = self._fbs_iterative(formula, steps)
        
        if not isinstance(formula, Atom) and formula.is_dnf():
            formula = espresso_exprs(formula)[0]

        if 'time' in sys.argv:
            print(time.time() - start)

        return formula

    def _fbs_iterative(self, formula, step):
        stack = [_fbs_iterate_item(
                formula=formula,
                parent=None,
                step=step,
                inv_nf=False)]
        
        while True:
            item = stack.pop()
            formula = item.formula

            if isinstance(formula, Constant):
                result = formula
            
            
            elif isinstance(formula, Variable):
                symbol = formula.name
                step = item.step

                if not step or not item.remained:
                    if (step, symbol) in self.cg:
                        result = ONE
                    elif (step, symbol) in self.cng:
                        result = ZERO
                    else:
                        result = var('{}_{}'.format(symbol, step))
                    
                elif item.remained:
                    item.remained = 1

                    stack.append(_fbs_iterate_item(
                        formula=self.cause(symbol),
                        parent=item,
                        step=step-1,
                        inv_nf=item.inv_nf))
                    continue
                    
                if step > 0:
                    result = Or(result, item.childs[0])


            elif isinstance(formula, Complement):
                if item.remained:
                    item.remained = 1
                    
                    stack.append(_fbs_iterate_item(
                        formula=Not(formula),
                        parent=item,
                        step=item.step,
                        inv_nf=not item.inv_nf))
                    continue
                
                else:
                    result = Not(item.childs[0])


            elif isinstance(formula, OrAndOp):
                Op = And if isinstance(formula, AndOp) else Or

                if item.remained:
                    item.remained = 2

                    stack.append(_fbs_iterate_item(
                        formula=formula.xs[0],
                        parent=item,
                        step=item.step,
                        inv_nf=item.inv_nf))
                    stack.append(_fbs_iterate_item(
                        formula=Op(*formula.xs[1:]),
                        parent=item,
                        step=item.step,
                        inv_nf=item.inv_nf))
                    continue
                
                else:
                    result = Op(item.childs[0], item.childs[1])


            result = result.to_cnf() if item.inv_nf else result.to_dnf()


            if item.parent == None:
                break
            else:
                item.parent.remained -= 1
                item.parent.childs.append(result)
                if not item.parent.remained:
                    stack.append(item.parent)
                continue


        return result
        
    
    def _fbs(self, formula, i, inv_nf=False):
        if isinstance(formula, Constant):
            pass

        elif isinstance(formula, Variable):
            symbol = formula.name
            if (i,symbol) in self.cg:
                return True
            elif (i,symbol) in self.cng:
                formula = ZERO
            else:
                formula = var('{}_{}'.format(symbol, i))
            
            if i > 0:
                formula = Or(formula, self._fbs(self.cause(symbol), i-1, inv_nf))

        elif isinstance(formula, Complement):
            formula = Not(self._fbs(Not(formula), i, not inv_nf))

        elif isinstance(formula, OrAndOp):
            Op = And if isinstance(formula, AndOp) else Or
            formula = Op(
                self._fbs(formula.xs[0], i, inv_nf),
                self._fbs(Op(*formula.xs[1:]), i, inv_nf))

        else:
            assert()

        formula = formula.to_cnf() if inv_nf else formula.to_dnf()

        return formula


    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return self._causes == other._causes

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(self._causes)

    def __str__(self):
        return 'ReactionSystem({})'.format(str(self._causes))

    def __repr__(self):
        return str(self)



from setuptools import setup, find_packages
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

name = 'fbp_calculator'

github = 'https://github.com/deselmo/FBP_Calculator'

with open(path.join(here, name, 'release.py')) as f:
    # Defines __version__
    exec(f.read())

setup(
    name=name,
    version=__version__, # pylint: disable=E0602
    description='FBP Calculator is a Python tool to calculate predicor for Reaction System.',
    url=github,
    author='William Guglielmo',
    author_email='william@deselmo.com',
    packages=find_packages(),
    install_requires=[
        'pyeda>=0.28.0',
        'PyQt5>=5.10.1',
        'XlsxWriter>=1.0.2'
    ],
    entry_points={
        'console_scripts': [
            '{}={}:main'.format(name, name),
        ],
    },
    project_urls={
        'Bug Reports': '{}/issues'.format(github),
        'Source': '{}/'.format(github),
    },
    license='MIT'
)


# -*- coding: utf-8 -*-

import re
import xlsxwriter
from PyQt5 import QtCore, QtGui, QtWidgets
from .ui_dialogfbp import Ui_DialogFBP
from .calculatorfbp import QThreadCalculatorFBP
from .reaction_adapter import reaction_invadapter

class DialogFBP(QtWidgets.QDialog, Ui_DialogFBP):
    def __init__(self, parent,
            symbols, steps, reaction_set, context_given_set, context_not_given_set):
        super(DialogFBP, self).__init__(parent)
        self.setupUi(self)

        self.setAttribute(QtCore.Qt.WA_DeleteOnClose)

        self.formulaType_defaultIndex = 1
        self.comboBoxFormulaType.setCurrentIndex(self.formulaType_defaultIndex)
        self.textBrowserFormula.setVisible(self.formulaType_defaultIndex == 0)
        self.listFormula.setVisible(self.formulaType_defaultIndex == 1)
        self.tableWidgetFormula.setVisible(self.formulaType_defaultIndex == 2)

        self.symbols = symbols
        self.steps = steps
        self.reaction_set = reaction_set
        self.context_given_set = context_given_set
        self.context_not_given_set = context_not_given_set

        self.lineEditSymbols.setText(reaction_invadapter(self.symbols))
        self.lineEditSteps.setText(str(self.steps))
        
        self.labelLoadingImage.setMovie(QtGui.QMovie(":/loader.gif"))
        self.labelLoadingImage.movie().start()

        self.toolButtonSave.setVisible(False)
        self.toolButtonSave.clicked.connect(self.toolButtonSave_clicked)

        self.comboBoxFormulaType.currentIndexChanged.connect(self.comboBoxFormulaType_currentIndexChanged)

        self.listFormula.verticalScrollBar().valueChanged.connect(self.listFormula_scrollBar_valueChanged)
        self.tableWidgetFormula.horizontalScrollBar().valueChanged.connect(self.tableWidgetFormula_horizontalScrollBar_valueChanged)
        self.tableWidgetFormula.verticalScrollBar().valueChanged.connect(self.tableWidgetFormula_verticalScrollBar_valueChanged)


        self.QThreadCalculatorFBP = QThreadCalculatorFBP(self)
        self.QThreadCalculatorFBP.finished.connect(self.QThread_finishedCalculatorFBP)
        self.QThreadCalculatorFBP.start()

    def toolButtonSave_clicked(self):
        formulaType_index = self.comboBoxFormulaType.currentIndex()
        if formulaType_index == 0:
            file_name, _ = QtWidgets.QFileDialog.getSaveFileName(self, 
                'Save result as', 
                'untitled.txt',
                'TXT files (*.txt)')

            if not file_name:
                return
            try:
                with open(file_name, 'w') as file:
                    file.write(self.save_text)
            except Exception as e:
                error_message = str(e)
                if 'Errno' in error_message:
                    error_message = error_message.split('] ')[1]
                QtWidgets.QMessageBox.critical(self,
                    'Error when saving the file',
                    '{}'.format(error_message),
                    QtWidgets.QMessageBox.Close,
                    QtWidgets.QMessageBox.Close)

        elif formulaType_index == 1:
            file_name, _ = QtWidgets.QFileDialog.getSaveFileName(self, 
                'Save result as', 
                'untitled.txt',
                'TXT files (*.txt)')
            
            if not file_name:
                return
            try:
                with open(file_name, 'w') as file:
                    file.write(self.save_list)
            except Exception as e:
                error_message = str(e)
                if 'Errno' in error_message:
                    error_message = error_message.split('] ')[1]
                QtWidgets.QMessageBox.critical(self,
                    'Error when saving the file',
                    '{}'.format(error_message),
                    QtWidgets.QMessageBox.Close,
                    QtWidgets.QMessageBox.Close)

        elif formulaType_index == 2:
            file_name, _ = QtWidgets.QFileDialog.getSaveFileName(self, 
                'Save result as', 
                'untitled.xlsx',
                'XLSX files (*.xlsx)')
            
            if not file_name:
                return
            workbook = xlsxwriter.Workbook(file_name)
            worksheet = workbook.add_worksheet()

            if isinstance(self.formula, bool):
                worksheet.write(0, 0, str(self.formula))

            else:
                for i in range(0, self.steps):
                    worksheet.write(0, i, str(i+1))

                for i in range(0, len(self.formula_table)):
                    row = self.formula_table[i]
                    for column in row:
                        text = row[column]
                        worksheet.write(i+1, column, text)

            try:
                workbook.close()
            except Exception as e:
                error_message = str(e)
                if 'Errno' in error_message:
                    error_message = error_message.split('] ')[1]
                QtWidgets.QMessageBox.critical(self,
                    'Error when saving the file',
                    '{}'.format(error_message),
                    QtWidgets.QMessageBox.Close,
                    QtWidgets.QMessageBox.Close)
            



    def resizeEvent(self, event):
        self.listFormula_fillSpace()
        self.tableWidgetFormula_fillVerticalSpace()
        self.tableWidgetFormula_fillHorizontalSpace()

    def closeEvent(self, event):
        self.QThreadCalculatorFBP.stop()
        self.QThreadCalculatorFBP.wait()
        event.accept()
    
    def QThread_finishedCalculatorFBP(self):
        if self.QThreadCalculatorFBP.stopped:
            return

        self.labelLoadingImage.setVisible(False)
        self.labelLoadingImage.movie().stop()
        self.labelComputing.setVisible(False)

        if not self.QThreadCalculatorFBP.result['completed']:
            self.labelComputing.setStyleSheet("QLabel { color : red; font-weight:600; }")
            self.labelComputing.setText('Error during the fbp calculation')
            self.labelComputing.setVisible(True)
            return
        
        self.formula = self.QThreadCalculatorFBP.result['formula']
        self.formula_table = self.QThreadCalculatorFBP.result['formula_table']

        self.toolButtonSave.setVisible(True)

        self.comboBoxFormulaType.setEnabled(True)
        self.comboBoxFormulaType_currentIndexChanged(self.formulaType_defaultIndex)

        self.raise_()


    def comboBoxFormulaType_currentIndexChanged(self, index):
        if index == 0:
            self.textBrowserFormula_show()
        elif index == 1:
            self.listFormula_show()
        elif index == 2:
            self.tableWidgetFormula_show()


    def textBrowserFormula_show(self):
        self.listFormula.setVisible(False)
        self.tableWidgetFormula.setVisible(False)
        self.textBrowserFormula.setVisible(True)

        if not self.textBrowserFormula.isEnabled():
            self.textBrowserFormula.setEnabled(True)
            self.textBrowserFormula_initialize()
            
    def textBrowserFormula_initialize(self):
        if isinstance(self.formula, bool):
            text = str(self.formula)
            self.textBrowserFormula.setText(text)
            self.save_text = text
            return

        text_subbed = ''
        prebrackets = len(self.formula) > 1
        for i in range(0, len(self.formula)):
            if i > 0: text_subbed += ' ∨ '
            backets = prebrackets and len(self.formula[i]) > 1
            if backets: text_subbed += '('
            for j in range(0, len(self.formula[i])):
                if j > 0: text_subbed += ' ∧ '
                n, s = self.formula[i][j]
                text_subbed += '{}<sub>{}</sub>'.format(s, str(n))
            if backets: text_subbed += ')'
        self.textBrowserFormula.setText(text_subbed)
        save_text = text_subbed
        save_text = re.sub('<sub>', '_', save_text)
        save_text = re.sub('</sub>', '', save_text)
        self.save_text = save_text


    def listFormula_show(self):
        self.textBrowserFormula.setVisible(False)
        self.tableWidgetFormula.setVisible(False)
        self.listFormula.setVisible(True)

        if not self.listFormula.isEnabled():
            self.listFormula.setEnabled(True)
            self.listFormula_initialize()
        else:
            self.listFormula_fillSpace()
            
    def listFormula_initialize(self):
        if isinstance(self.formula, bool):
            text = str(self.formula)
            self.listFormula.addItem(QtWidgets.QListWidgetItem(text))
            self.save_list = text
            return
        
        text = ''
        for f in self.formula:
            for i in range(0, len(f)):
                n, s = f[i]
                text += '{}_{} '.format(s, str(n))
            text = text[:-1]
            text += '\r\n'
        self.save_list = text

        self.listFormula_fillSpace()

    def listFormula_fillSpace(self):
        if (not self.listFormula.isEnabled() or not self.listFormula.isVisible()
                or isinstance(self.formula, bool)):
            return

        for _ in range(self.listFormula.count(), len(self.formula)):
            if self.listFormula.verticalScrollBar().maximum() != 0:
                break
            self.listFormula_addRow()
    
    def listFormula_scrollBar_valueChanged(self, value):
        if (not self.listFormula.isEnabled() or not self.listFormula.isVisible()
                or isinstance(self.formula, bool)):
            return

        listFormula_len = self.listFormula.count()

        if (listFormula_len < len(self.formula) and
                value == self.listFormula.verticalScrollBar().maximum()):
            self.listFormula_addRow()

    def listFormula_addRow(self):
        f = self.formula[self.listFormula.count()]
        text = ''
        for i in range(0, len(f)):
            n, s = f[i]
            text += '{}<sub>{}</sub> '.format(s, str(n))
        text = text[:-1]

        label = QtWidgets.QLabel(text)
        label.setContentsMargins(4,4,4,4)
        item = QtWidgets.QListWidgetItem()
        item.setSizeHint(QtCore.QSize(0, label.sizeHint().height()+4))
        self.listFormula.addItem(item)
        self.listFormula.setItemWidget(item, label)


    def tableWidgetFormula_show(self):
        self.textBrowserFormula.setVisible(False)
        self.listFormula.setVisible(False)
        self.tableWidgetFormula.setVisible(True)

        if not self.tableWidgetFormula.isEnabled():
            self.tableWidgetFormula.setEnabled(True)
            self.tableWidgetFormula_initialize()
        else:
            self.tableWidgetFormula_fillVerticalSpace()
            self.tableWidgetFormula_fillHorizontalSpace()

    def tableWidgetFormula_initialize(self):
        if isinstance(self.formula, bool):
            self.tableWidgetFormula.horizontalHeader().setVisible(False)
            self.tableWidgetFormula.setRowCount(1)
            self.tableWidgetFormula.setColumnCount(1)
            self.tableWidgetFormula_addCell(0, 0, str(self.formula))

            self.tableWidgetFormula_resizeToContent()
            return

        self.tableWidgetFormula_fillHorizontalSpace()
        self.tableWidgetFormula_fillVerticalSpace()
        
    def tableWidgetFormula_fillVerticalSpace(self):
        if (not self.tableWidgetFormula.isEnabled() or not self.tableWidgetFormula.isVisible()
                or isinstance(self.formula, bool)):
            return
        
        for _ in range(self.tableWidgetFormula.rowCount(), len(self.formula_table)):
            if self.tableWidgetFormula.verticalScrollBar().maximum() != 0:
                break
            self.tableWidgetFormula_addRow()

    def tableWidgetFormula_fillHorizontalSpace(self):
        if (not self.tableWidgetFormula.isEnabled() or not self.tableWidgetFormula.isVisible()
                or isinstance(self.formula, bool)):
            return
        
        for _ in range(self.tableWidgetFormula.columnCount(), self.steps):
            if self.tableWidgetFormula.horizontalScrollBar().maximum() != 0:
                break
            self.tableWidgetFormula_addColumn()

    def tableWidgetFormula_verticalScrollBar_valueChanged(self, value):
        if (not self.tableWidgetFormula.isEnabled() or not self.tableWidgetFormula.isVisible()
                or isinstance(self.formula, bool)):
            return

        if (self.tableWidgetFormula.rowCount() < len(self.formula_table) and
                value == self.tableWidgetFormula.verticalScrollBar().maximum()):
            self.tableWidgetFormula_addRow()

    def tableWidgetFormula_horizontalScrollBar_valueChanged(self, value):
        if (not self.tableWidgetFormula.isEnabled()) or isinstance(self.formula, bool):
            return

        if (self.tableWidgetFormula.columnCount() < self.steps and
                value == self.tableWidgetFormula.horizontalScrollBar().maximum()):
            self.tableWidgetFormula_addColumn()

    def tableWidgetFormula_addRow(self):
        column = self.tableWidgetFormula.columnCount()
        row = self.tableWidgetFormula.rowCount()
        self.tableWidgetFormula.setRowCount(row+1)
        
        f = self.formula_table[row]
        for i in range(0, column):
            if not i in f: continue
            s = f[i]
            self.tableWidgetFormula_addCell(row, i, s)

        self.tableWidgetFormula_resizeToContent()
    
    def tableWidgetFormula_addColumn(self):
        column = self.tableWidgetFormula.columnCount()
        row = self.tableWidgetFormula.rowCount()
        self.tableWidgetFormula.setColumnCount(column+1)

        self.tableWidgetFormula.setHorizontalHeaderItem(column, QtWidgets.QTableWidgetItem(str(column+1)))

        for i in range(0, row):
            f = self.formula_table[i]
            if not column in f: continue
            s = f[column]
            self.tableWidgetFormula_addCell(i, column, s)

        self.tableWidgetFormula_resizeToContent()

    def tableWidgetFormula_addCell(self, row, column, text):
        cellWidget = self.tableWidgetFormula.cellWidget(row, column)
        if cellWidget == None:
            label = QtWidgets.QLabel(text)
            label.setContentsMargins(8,2,8,2)
            label.setAlignment(QtCore.Qt.AlignLeft)
            self.tableWidgetFormula.setCellWidget(row, column, label)
        else:
            raise Exception('Add a cell more than one time')

    def tableWidgetFormula_resizeToContent(self):
        self.tableWidgetFormula.horizontalHeader().setResizeContentsPrecision(self.tableWidgetFormula.rowCount())
        self.tableWidgetFormula.resizeColumnsToContents()
        self.tableWidgetFormula.horizontalHeader().setSectionResizeMode(QtWidgets.QHeaderView.Fixed)
